// "The two-loop helicity amplitudes for qqb' -> V1V2 -> 4 leptons"
// T. Gehrmann, A. von Manteuffel, L. Tancredi
//
// qqvvamp package


#include "qqvvampprec.h"
#include <iostream>
#include <cmath>
#include <ginac/ginac.h>

#if defined(DOUBLE_PRECISION) && defined(QUAD_PRECISION)

using namespace std;

qqvvampprec::qqvvampprec(int Nf, IRscheme ir) :
  maxrel(1e-4), maxabs(1e-50), d(Nf,ir), q(Nf,ir)
#ifdef ARBITRARY_PRECISION
  , a(Nf,ir)
#endif
{  
}

void qqvvampprec::print_statistics(std::ostream& os) const {
  os << "number of events per number of Digits used:" << endl;
  map<long, long>::const_iterator i;
  for (i = eventsWithDigits.begin(); i != eventsWithDigits.end(); ++i) {
    os << "  " << i->first << " -> " << i->second << endl;
  }
  os << "\n";
}

// check if two results are compatible within the given precision
// and copy higher precision result to target array
template<class T1, class T2>
bool control_precision(T1 Eloprec[nEcoeff+1][ntype][nloop][nreim], 
		       T2 Ehiprec[nEcoeff+1][ntype][nloop][nreim],
		       double Eresult[nEcoeff+1][ntype][nloop][nreim],
		       double maxrel, double maxabs) {
  bool stable = true;
  for (int i = 1; i <= nEcoeff; ++i) {
    for (int j = 0; j < ntype; ++j) {
      for (int l = 0; l < nloop; ++l) {
	for (int r = 0 ; r < nreim; ++r) {
	  double lo = todouble(Eloprec[i][j][l][r]);
	  double hi = todouble(Ehiprec[i][j][l][r]);
	  if (fabs(lo-hi) > maxabs && fabs(lo-hi) > maxrel*(fabs(lo)+fabs(hi))) {
	    stable = false;
	  }
	  Eresult[i][j][l][r] = hi;
	}
      }
    }
  }
  return stable;
}

void qqvvampprec::compute(double s, double t, double ma2, double mb2) {
  using GiNaC::Digits;
  long maxdigits = 500;
  bool recycle_gpls = true;
  
  // run1: double precision GPL, quad precision algebra
  Digits = 17;
  q.compute(s, t, ma2, mb2);
  
  // run2: double precision algebra (recycling previous GPL evaluations)
  if (recycle_gpls) {
    size_t hi = qqvvamp<double>::rstartalgebraic;
    for (size_t i = 0 ; i <= hi; ++i)
      d.r[i] = todouble(q.r[i]);
  }
  d.compute_gpls = !recycle_gpls;
  d.compute(s, t, ma2, mb2);
  if (control_precision(d.E, q.E, E, maxrel, maxabs)) {
    ++eventsWithDigits[Digits];
    return;
  }

  // run3: quad precision GPL, quad precision algebra
  Digits = 32;
  q.compute(s, t, ma2, mb2);
  if (control_precision(d.E, q.E, E, maxrel, maxabs)) {
    cerr << "qqvvamp: increasing precision to " << Digits << endl;
    ++eventsWithDigits[Digits];
    return;
  }

#ifdef ARBITRARY_PRECISION
  // run4 and more: arbitrary precision GPL and algebra
  for (Digits = 64; Digits <= maxdigits; Digits = Digits*2) {
    cerr << "qqvvamp: increasing precision to " << Digits << endl;
    cln::cl_F (*n)(const GiNaC::ex&) = &qqvvamp<cln::cl_F>::n;
    a.compute(n(s), n(t), n(ma2), n(mb2));
    if (control_precision(E, a.E, E, maxrel, maxabs)) {
      ++eventsWithDigits[Digits];
      return;
    }
  }
#endif

  // no convergence
  throw runtime_error("qqvvamp: could not converge to stable result !");
}

#endif

#include "qqvvamp.xi"
